@@ -12,8 +12,10 @@ import sys
 import re
 import pysrt
 import delegator
+from datetime import datetime
 from subliminal import *
 from babelfish import Language
+from collections import OrderedDict
 
 try:
     from cleanvid.caselessdictionary import CaselessDictionary
@@ -24,10 +26,13 @@ from itertools import tee
 __script_location__ = os.path.dirname(os.path.realpath(__file__))
 
 VIDEO_DEFAULT_PARAMS = '-c:v libx264 -preset slow -crf 22'
-AUDIO_DEFAULT_PARAMS = '-c:a aac -ac 2 -ab 224k -ar 44100'
+AUDIO_DEFAULT_PARAMS = '-c:a aac -ab 224k -ar 44100'
+# for downmixing, https://superuser.com/questions/852400 was helpful
+AUDIO_DOWNMIX_FILTER = 'pan=stereo|FL=0.8*FC + 0.6*FL + 0.6*BL + 0.5*LFE|FR=0.8*FC + 0.6*FR + 0.6*BR + 0.5*LFE'
 SUBTITLE_DEFAULT_LANG = 'eng'
 PLEX_AUTO_SKIP_DEFAULT_CONFIG = '{"markers":{},"offsets":{},"tags":{},"allowed":{"users":[],"clients":[],"keys":[]},"blocked":{"users":[],"clients":[],"keys":[]},"clients":{},"mode":{}}'
 
+
 # thanks https://docs.python.org/3/library/itertools.html#recipes
 def pairwise(iterable):
     a, b = tee(iterable)
@@ -39,38 +44,80 @@ def pairwise(iterable):
 def GetFormatAndStreamInfo(vidFileSpec):
     result = None
     if os.path.isfile(vidFileSpec):
-        ffprobeCmd = "ffprobe -v quiet -print_format json -show_format -show_streams \"" + vidFileSpec + "\""
+        ffprobeCmd = "ffprobe -loglevel quiet -print_format json -show_format -show_streams \"" + vidFileSpec + "\""
         ffprobeResult = delegator.run(ffprobeCmd, block=True)
         if ffprobeResult.return_code == 0:
             result = json.loads(ffprobeResult.out)
     return result
 
 
+######## GetStreamSubtitleMap ###############################################
+def GetStreamSubtitleMap(vidFileSpec):
+    result = None
+    if os.path.isfile(vidFileSpec):
+        ffprobeCmd = (
+            "ffprobe -loglevel quiet -select_streams s -show_entries stream=index:stream_tags=language -of csv=p=0 \""
+            + vidFileSpec
+            + "\""
+        )
+        ffprobeResult = delegator.run(ffprobeCmd, block=True)
+        if ffprobeResult.return_code == 0:
+            # e.g. for ara and chi, "-map 0:5 -map 0:7" or "-map 0:s:3 -map 0:s:5"
+            # 2,eng
+            # 3,eng
+            # 4,eng
+            # 5,ara
+            # 6,bul
+            # 7,chi
+            # 8,cze
+            # 9,dan
+            result = OrderedDict()
+            for l in [x.split(',') for x in ffprobeResult.out.split()]:
+                result[int(l[0])] = l[1]
+    return result
+
+
+######## HasAudioMoreThanStereo ###############################################
+def HasAudioMoreThanStereo(vidFileSpec):
+    result = False
+    if os.path.isfile(vidFileSpec):
+        ffprobeCmd = (
+            "ffprobe -loglevel quiet -select_streams a -show_entries stream=channels -of csv=p=0 \""
+            + vidFileSpec
+            + "\""
+        )
+        ffprobeResult = delegator.run(ffprobeCmd, block=True)
+        if ffprobeResult.return_code == 0:
+            result = any([int(x) for x in list(set(ffprobeResult.out.split())) if int(x) > 2])
+    return result
+
+
+######## SplitLanguageIfForced #####################################################
+def SplitLanguageIfForced(lang):
+    srtLanguageSplit = lang.split(':')
+    srtLanguage = srtLanguageSplit[0]
+    srtForceIndex = int(srtLanguageSplit[1]) if len(srtLanguageSplit) > 1 else None
+    return srtLanguage, srtForceIndex
+
+
 ######## ExtractSubtitles #####################################################
 def ExtractSubtitles(vidFileSpec, srtLanguage):
     subFileSpec = ""
-    if (
-        (streamInfo := GetFormatAndStreamInfo(vidFileSpec))
-        and ('streams' in streamInfo)
-        and (len(streamInfo['streams']) > 0)
-        and (
-            streams := [
-                x
-                for x in streamInfo['streams']
-                if ('codec_type' in x)
-                and ('index' in x)
-                and (x['codec_type'] == 'subtitle')
-                and ('codec_name' in x)
-                and (x['codec_name'] == 'subrip')
-                and ('tags' in x)
-                and ('language' in x['tags'])
-                and (x['tags']['language'] == srtLanguage)
-            ]
-        )
+    srtLanguage, srtForceIndex = SplitLanguageIfForced(srtLanguage)
+    if (streamInfo := GetStreamSubtitleMap(vidFileSpec)) and (
+        stream := next(iter([k for k, v in streamInfo.items() if (v == srtLanguage)]), None)
+        if not srtForceIndex
+        else srtForceIndex
     ):
         subFileParts = os.path.splitext(vidFileSpec)
         subFileSpec = subFileParts[0] + "." + srtLanguage + ".srt"
-        ffmpegCmd = "ffmpeg -y -i \"" + vidFileSpec + f"\" -map 0:{streams[0]['index']} \"" + subFileSpec + "\""
+        ffmpegCmd = (
+            "ffmpeg -hide_banner -nostats -loglevel error -y -i \""
+            + vidFileSpec
+            + f"\" -map 0:{stream} \""
+            + subFileSpec
+            + "\""
+        )
         ffmpegResult = delegator.run(ffmpegCmd, block=True)
         if (ffmpegResult.return_code != 0) or (not os.path.isfile(subFileSpec)):
             subFileSpec = ""
@@ -86,6 +133,7 @@ def GetSubtitles(vidFileSpec, srtLanguage, offline=False):
         else:
             if os.path.isfile(vidFileSpec):
                 subFileParts = os.path.splitext(vidFileSpec)
+                srtLanguage, srtForceIndex = SplitLanguageIfForced(srtLanguage)
                 subFileSpec = subFileParts[0] + "." + str(Language(srtLanguage)) + ".srt"
                 if not os.path.isfile(subFileSpec):
                     video = Video.fromname(vidFileSpec)
@@ -101,7 +149,6 @@ def GetSubtitles(vidFileSpec, srtLanguage, offline=False):
 ######## UTF8Convert #########################################################
 # attempt to convert any text file to UTF-* without BOM and normalize line endings
 def UTF8Convert(fileSpec, universalEndline=True):
-
     # Read from file
     with open(fileSpec, 'rb') as f:
         raw = f.read()
@@ -131,6 +178,7 @@ class VidCleaner(object):
     inputSubsFileSpec = ""
     cleanSubsFileSpec = ""
     edlFileSpec = ""
+    jsonFileSpec = ""
     tmpSubsFileSpec = ""
     assSubsFileSpec = ""
     outputVidFileSpec = ""
@@ -146,10 +194,12 @@ class VidCleaner(object):
     subsLang = SUBTITLE_DEFAULT_LANG
     vParams = VIDEO_DEFAULT_PARAMS
     aParams = AUDIO_DEFAULT_PARAMS
+    aDownmix = False
     plexAutoSkipJson = ""
     plexAutoSkipId = ""
     swearsMap = CaselessDictionary({})
     muteTimeList = []
+    jsonDumpList = None
 
     ######## init #################################################################
 
@@ -160,20 +210,21 @@ class VidCleaner(object):
         oVidFileSpec,
         oSubsFileSpec,
         iSwearsFileSpec,
-        swearsPadSec=0,
+        swearsPadSec=0.0,
         embedSubs=False,
         fullSubs=False,
         subsOnly=False,
         edl=False,
+        jsonDump=False,
         subsLang=SUBTITLE_DEFAULT_LANG,
         reEncode=False,
         hardCode=False,
         vParams=VIDEO_DEFAULT_PARAMS,
         aParams=AUDIO_DEFAULT_PARAMS,
+        aDownmix=False,
         plexAutoSkipJson="",
         plexAutoSkipId="",
     ):
-
         if (iVidFileSpec is not None) and os.path.isfile(iVidFileSpec):
             self.inputVidFileSpec = iVidFileSpec
         else:
@@ -197,11 +248,12 @@ class VidCleaner(object):
             if os.path.isfile(self.cleanSubsFileSpec):
                 os.remove(self.cleanSubsFileSpec)
 
-        self.swearsPadMillisec = swearsPadSec * 1000
+        self.swearsPadMillisec = round(swearsPadSec * 1000.0)
         self.embedSubs = embedSubs
         self.fullSubs = fullSubs
         self.subsOnly = subsOnly or edl or (plexAutoSkipJson and plexAutoSkipId)
         self.edl = edl
+        self.jsonDumpList = [] if jsonDump else None
         self.plexAutoSkipJson = plexAutoSkipJson
         self.plexAutoSkipId = plexAutoSkipId
         self.reEncode = reEncode
@@ -209,6 +261,7 @@ class VidCleaner(object):
         self.subsLang = subsLang
         self.vParams = vParams
         self.aParams = aParams
+        self.aDownmix = aDownmix
         if self.vParams.startswith('base64:'):
             self.vParams = base64.b64decode(self.vParams[7:]).decode('utf-8')
         if self.aParams.startswith('base64:'):
@@ -221,6 +274,8 @@ class VidCleaner(object):
                 os.remove(self.cleanSubsFileSpec)
             if os.path.isfile(self.edlFileSpec):
                 os.remove(self.edlFileSpec)
+            if os.path.isfile(self.jsonFileSpec):
+                os.remove(self.jsonFileSpec)
         if os.path.isfile(self.tmpSubsFileSpec):
             os.remove(self.tmpSubsFileSpec)
         if os.path.isfile(self.assSubsFileSpec):
@@ -248,6 +303,10 @@ class VidCleaner(object):
             cleanSubFileParts = os.path.splitext(self.cleanSubsFileSpec)
             self.edlFileSpec = cleanSubFileParts[0] + '.edl'
 
+        if (self.jsonDumpList is not None) and (not self.jsonFileSpec):
+            cleanSubFileParts = os.path.splitext(self.cleanSubsFileSpec)
+            self.jsonFileSpec = cleanSubFileParts[0] + '.json'
+
         lines = []
 
         with open(self.swearsFileSpec) as f:
@@ -266,6 +325,16 @@ class VidCleaner(object):
         newSubs = pysrt.SubRipFile()
         newTimestampPairs = []
 
+        # append a dummy sub at the very end so that pairwise can peek and see nothing
+        subs.append(
+            pysrt.SubRipItem(
+                index=len(subs) + 1,
+                start=(subs[-1].end.seconds if subs else 0) + 1,
+                end=(subs[-1].end.seconds if subs else 0) + 2,
+                text='Fin',
+            )
+        )
+
         # for each subtitle in the set
         # if text contains profanity...
         # OR if the next text contains profanity and lies within the pad ...
@@ -302,6 +371,15 @@ class VidCleaner(object):
                 )
             ):
                 subScrubbed = newText != sub.text
+                if subScrubbed and (self.jsonDumpList is not None):
+                    self.jsonDumpList.append(
+                        {
+                            'old': sub.text,
+                            'new': newText,
+                            'start': str(sub.start),
+                            'end': str(sub.end),
+                        }
+                    )
                 newSub = sub
                 newSub.text = newText
                 newSubs.append(newSub)
@@ -321,6 +399,26 @@ class VidCleaner(object):
                 prevNaughtySub = None
 
         newSubs.save(self.cleanSubsFileSpec)
+        if self.jsonDumpList is not None:
+            with open(self.jsonFileSpec, "w") as f:
+                f.write(
+                    json.dumps(
+                        {
+                            "now": datetime.now().isoformat(),
+                            "edits": self.jsonDumpList,
+                            "media": {
+                                "input": self.inputVidFileSpec,
+                                "output": self.outputVidFileSpec,
+                                "ffprobe": GetFormatAndStreamInfo(self.inputVidFileSpec),
+                            },
+                            "subtitles": {
+                                "input": self.inputSubsFileSpec,
+                                "output": self.cleanSubsFileSpec,
+                            },
+                        },
+                        indent=4,
+                    )
+                )
 
         self.muteTimeList = []
         edlLines = []
@@ -330,7 +428,7 @@ class VidCleaner(object):
             plexDict["markers"][self.plexAutoSkipId] = []
             plexDict["mode"][self.plexAutoSkipId] = "volume"
 
-        for timePair in newTimestampPairs:
+        for timePair, timePairPeek in pairwise(newTimestampPairs):
             lineStart = (
                 (timePair[0].hour * 60.0 * 60.0)
                 + (timePair[0].minute * 60.0)
@@ -343,8 +441,29 @@ class VidCleaner(object):
                 + timePair[1].second
                 + (timePair[1].microsecond / 1000000.0)
             )
+            lineStartPeek = (
+                (timePairPeek[0].hour * 60.0 * 60.0)
+                + (timePairPeek[0].minute * 60.0)
+                + timePairPeek[0].second
+                + (timePairPeek[0].microsecond / 1000000.0)
+            )
             self.muteTimeList.append(
-                "volume=enable='between(t," + format(lineStart, '.3f') + "," + format(lineEnd, '.3f') + ")':volume=0"
+                "afade=enable='between(t,"
+                + format(lineStart, '.3f')
+                + ","
+                + format(lineEnd, '.3f')
+                + ")':t=out:st="
+                + format(lineStart, '.3f')
+                + ":d=10ms"
+            )
+            self.muteTimeList.append(
+                "afade=enable='between(t,"
+                + format(lineEnd, '.3f')
+                + ","
+                + format(lineStartPeek, '.3f')
+                + ")':t=in:st="
+                + format(lineEnd, '.3f')
+                + ":d=10ms"
             )
             if self.edl:
                 edlLines.append(f"{format(lineStart, '.1f')}\t{format(lineEnd, '.3f')}\t1")
@@ -365,7 +484,6 @@ class VidCleaner(object):
 
     ######## MultiplexCleanVideo ###################################################
     def MultiplexCleanVideo(self):
-
         # if we're don't *have* to generate a new video file, don't
         # we need to generate a video file if any of the following are true:
         # - we were explicitly asked to re-encode
@@ -376,7 +494,7 @@ class VidCleaner(object):
             if self.reEncode or self.hardCode:
                 if self.hardCode and os.path.isfile(self.cleanSubsFileSpec):
                     self.assSubsFileSpec = self.cleanSubsFileSpec + '.ass'
-                    subConvCmd = f"ffmpeg -y -i {self.cleanSubsFileSpec} {self.assSubsFileSpec}"
+                    subConvCmd = f"ffmpeg -hide_banner -nostats -loglevel error -y -i {self.cleanSubsFileSpec} {self.assSubsFileSpec}"
                     subConvResult = delegator.run(subConvCmd, block=True)
                     if (subConvResult.return_code == 0) and os.path.isfile(self.assSubsFileSpec):
                         videoArgs = f"{self.vParams} -vf \"ass={self.assSubsFileSpec}\""
@@ -388,22 +506,24 @@ class VidCleaner(object):
                     videoArgs = self.vParams
             else:
                 videoArgs = "-c:v copy"
+            if self.aDownmix and HasAudioMoreThanStereo(self.inputVidFileSpec):
+                self.muteTimeList.insert(0, AUDIO_DOWNMIX_FILTER)
             if (not self.subsOnly) and (len(self.muteTimeList) > 0):
-                audioArgs = " -af \"" + ",".join(self.muteTimeList) + "\" "
+                audioFilter = " -af \"" + ",".join(self.muteTimeList) + "\" "
             else:
-                audioArgs = " "
+                audioFilter = " "
             if self.embedSubs and os.path.isfile(self.cleanSubsFileSpec):
                 outFileParts = os.path.splitext(self.outputVidFileSpec)
                 subsArgs = f" -i \"{self.cleanSubsFileSpec}\" -map 0 -map -0:s -map 1 -c:s {'mov_text' if outFileParts[1] == '.mp4' else 'srt'} -disposition:s:0 default -metadata:s:s:0 language={self.subsLang} "
             else:
                 subsArgs = " -sn "
             ffmpegCmd = (
-                "ffmpeg -y -i \""
+                "ffmpeg -hide_banner -nostats -loglevel error -y -i \""
                 + self.inputVidFileSpec
                 + "\""
                 + subsArgs
                 + videoArgs
-                + audioArgs
+                + audioFilter
                 + f"{self.aParams} \""
                 + self.outputVidFileSpec
                 + "\""
@@ -451,12 +571,12 @@ def RunCleanvid():
     parser.add_argument(
         '-l',
         '--lang',
-        help=f'language for srt download (default is "{SUBTITLE_DEFAULT_LANG}")',
+        help=f'language for extracting srt from video file or srt download (default is "{SUBTITLE_DEFAULT_LANG}")',
         default=SUBTITLE_DEFAULT_LANG,
         metavar='<language>',
     )
     parser.add_argument(
-        '-p', '--pad', help='pad (seconds) around profanity', metavar='<int>', dest="pad", type=int, default=0
+        '-p', '--pad', help='pad (seconds) around profanity', metavar='<int>', dest="pad", type=float, default=0.0
     )
     parser.add_argument(
         '-e',
@@ -490,6 +610,12 @@ def RunCleanvid():
         dest='edl',
         action='store_true',
     )
+    parser.add_argument(
+        '--json',
+        help='generate JSON file with muted subtitles and their contents',
+        dest='json',
+        action='store_true',
+    )
     parser.add_argument('-r', '--re-encode', help='Re-encode video', dest='reEncode', action='store_true')
     parser.add_argument(
         '-b', '--burn', help='Hard-coded subtitles (implies re-encode)', dest='hardCode', action='store_true'
@@ -504,6 +630,9 @@ def RunCleanvid():
     parser.add_argument(
         '-a', '--audio-params', help='Audio parameters for ffmpeg', dest='aParams', default=AUDIO_DEFAULT_PARAMS
     )
+    parser.add_argument(
+        '-d', '--downmix', help='Downmix to stereo (if not already stereo)', dest='aDownmix', action='store_true'
+    )
     parser.set_defaults(
         embedSubs=False,
         fullSubs=False,
@@ -545,11 +674,13 @@ def RunCleanvid():
         args.fullSubs,
         args.subsOnly,
         args.edl,
+        args.json,
         lang,
         args.reEncode,
         args.hardCode,
         args.vParams,
         args.aParams,
+        args.aDownmix,
         plexFile,
         args.plexAutoSkipId,
     )
